@@ -1,13 +1,16 @@
 use core::convert::TryInto;
+use core::fmt::{self, Debug};
+use core::mem::ManuallyDrop;
 use core::ptr;
 use core::slice;
 use core::time;
-use std::mem::ManuallyDrop;
 
 extern crate alloc;
-use alloc::{borrow::Cow, sync::Arc};
+use alloc::borrow::Cow;
+use alloc::boxed::Box;
+use alloc::sync::Arc;
 
-use embedded_svc::{errors, mqtt::client};
+use embedded_svc::{errors, mqtt::client, mqtt::client::Message};
 
 use esp_idf_hal::mutex::{Condvar, Mutex};
 
@@ -58,6 +61,9 @@ pub struct MqttClientConfiguration<'a> {
     pub buffer_size: usize,
     pub out_buffer_size: usize,
 
+    pub username: Option<&'a str>,
+    pub password: Option<&'a str>,
+
     pub use_global_ca_store: bool,
     pub skip_cert_common_name_check: bool,
     #[cfg(not(esp_idf_version = "4.3"))]
@@ -98,6 +104,9 @@ impl<'a> Default for MqttClientConfiguration<'a> {
             buffer_size: 0,
             out_buffer_size: 0,
 
+            username: None,
+            password: None,
+
             use_global_ca_store: false,
             skip_cert_common_name_check: false,
 
@@ -107,8 +116,8 @@ impl<'a> Default for MqttClientConfiguration<'a> {
     }
 }
 
-impl<'a> From<&MqttClientConfiguration<'a>> for (esp_mqtt_client_config_t, RawCstrs) {
-    fn from(conf: &MqttClientConfiguration<'a>) -> Self {
+impl<'a> From<&'a MqttClientConfiguration<'a>> for (esp_mqtt_client_config_t, RawCstrs) {
+    fn from(conf: &'a MqttClientConfiguration<'a>) -> Self {
         let mut cstrs = RawCstrs::new();
 
         let mut c_conf = esp_mqtt_client_config_t {
@@ -129,6 +138,9 @@ impl<'a> From<&MqttClientConfiguration<'a>> for (esp_mqtt_client_config_t, RawCs
             buffer_size: conf.buffer_size as _,
             out_buffer_size: conf.out_buffer_size as _,
 
+            username: cstrs.as_nptr(conf.username),
+            password: cstrs.as_nptr(conf.password),
+
             use_global_ca_store: conf.use_global_ca_store,
             skip_cert_common_name_check: conf.skip_cert_common_name_check,
             #[cfg(not(esp_idf_version = "4.3"))]
@@ -247,7 +259,7 @@ impl EspMqttClient {
     }
 
     fn new_with_raw_callback<'a>(
-        url: impl AsRef<str>,
+        url: impl AsRef<str> + 'a,
         conf: &'a MqttClientConfiguration<'a>,
         raw_callback: Box<dyn FnMut(esp_mqtt_event_handle_t)>,
     ) -> Result<Self, EspError>
@@ -309,7 +321,7 @@ impl EspMqttClient {
 impl EspMqttClient {
     pub fn new_async<'a, U, S>(
         url: S,
-        conf: &'a MqttClientConfiguration<'a>,
+        conf: /*TODO: issue with the compiler: &'a*/ MqttClientConfiguration<'a>,
     ) -> Result<
         (
             embedded_svc::utils::nonblocking::mqtt::client::AsyncClient<U, Mutex<Self>>,
@@ -323,7 +335,7 @@ impl EspMqttClient {
     >
     where
         U: embedded_svc::unblocker::nonblocking::Unblocker,
-        S: AsRef<str>,
+        S: AsRef<str> + 'a,
     {
         let connection = embedded_svc::utils::nonblocking::mqtt::client::AsyncConnection::<
             Condvar,
@@ -333,7 +345,7 @@ impl EspMqttClient {
 
         let cb_connection = connection.clone();
 
-        let client = Self::new_with_callback(url, conf, move |event| cb_connection.post(event))?;
+        let client = Self::new_with_callback(url, &conf, move |event| cb_connection.post(event))?;
 
         Ok((
             embedded_svc::utils::nonblocking::mqtt::client::AsyncClient::new(client),
@@ -506,6 +518,18 @@ impl<'a> EspMqttMessage<'a> {
     }
 }
 
+impl<'a> Debug for EspMqttMessage<'a> {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(
+            f,
+            "[id = {}, topic = {:?}, details = {:?}]",
+            self.id(),
+            self.retrieve_topic(),
+            self.details()
+        )
+    }
+}
+
 impl<'a> Drop for EspMqttMessage<'a> {
     fn drop(&mut self) {
         if let Some(state) = self.connection.as_ref() {
@@ -548,7 +572,7 @@ impl<'a> client::Message for EspMqttMessage<'a> {
     }
 }
 
-#[allow(suspicious_auto_trait_impls)]
+#[cfg_attr(version("1.61"), allow(suspicious_auto_trait_impls))]
 unsafe impl Send for Newtype<esp_mqtt_event_handle_t> {}
 
 struct EspMqttConnectionState {
